#include <BlynkSimpleEsp32_BLE.h>
#include <BLEDevice.h>
#include <BLEServer.h>
#include <ESP32Servo.h>
#include "Wire.h"
#include "Adafruit_PWMServoDriver.h"
Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver(0x40);

/* Comment this out to disable prints and save space */
//#define BLYNK_PRINT Serial

#define BLYNK_USE_DIRECT_CONNECT

#define JAWS_SERVO 12
#define NECK_SERVO 0
#define WINGS_SERVO 4
#define TAIL_SERVO 15
#define TEST_SERVO 8

#define MIN_DOWN 39
#define MIN_UP 36
#define MAX_DOWN 35
#define MAX_UP 34

// Define maximum and minimum number of "ticks" for the servo motors
// Range from 0 to 4095
// This determines the pulse width
#define MIN_PULSE_WIDTH 500 // this is the 'minimum' pulse length count (out of 4096)
#define MAX_PULSE_WIDTH 2600 // this is the 'maximum' pulse length count (out of 4096)
#define FREQUENCY 50

int testAngle = 90; //degrees
int testServo = TEST_SERVO;

bool wingsFlap = false;
unsigned int wingsAngle = 176;
bool incrementWings = false;
unsigned long lastWingUpdate = 176;

bool tailWag = false;
unsigned int tailAngle = 90;
bool incrementTail = false;
unsigned long lastTailUpdate = 0;

bool neckRotate = false;
unsigned int neckAngle = 78;
bool incrementNeck = false;
unsigned long lastNeckUpdate = 0;
unsigned int oldNeckAngle = 78;
unsigned int newNeckAngle = 78;
unsigned long lastNeckTurn = 0;

bool jawsRoar = false;
unsigned int jawAngle = 92;
unsigned long lastJawsUpdate = 0;
unsigned int oldJawAngle = 92;
unsigned int newJawAngle = 92;
unsigned long lastJawMove = 0;

// You should get Auth Token in the Blynk App.
// Go to the Project Settings (nut icon).
char auth[] = "YOUR_AUTH_CODE_HERE";

int pwToAnalogAng(int angle)
{
  int pulse_width, analog_value;
  if(angle < 0)
  {
    angle = 0;
  }
  else if(angle > 180)
  {
    angle = 180;
  }
  pulse_width = map(angle, 0, 180, MIN_PULSE_WIDTH, MAX_PULSE_WIDTH);
  analog_value = int(float(pulse_width) / 1000000 * FREQUENCY * 4096);
  return analog_value;
}

void flapWings()
{
  if(wingsFlap)
  {
    if(wingsAngle == 176)
    {
      incrementWings = false;
    }
    else if(wingsAngle == 64)
    {
      incrementWings = true;
    }
    if(millis() - lastWingUpdate > 50)
    {
      wingsAngle = (incrementWings) ? wingsAngle + 4: wingsAngle - 4;
      pwm.setPWM(WINGS_SERVO, 0, pwToAnalogAng(wingsAngle));
      lastWingUpdate = millis();
    }
  }
}

void rotateNeck()
{
  if(neckRotate)
  {
    if(millis() - lastNeckUpdate > 4000 && oldNeckAngle == newNeckAngle)
    {
      //Time to change neck direction
      if(newNeckAngle == 163)
      {
        newNeckAngle = 78;
        incrementNeck = false;
      }
      else if(newNeckAngle == 2)
      {
        newNeckAngle = 78;
        incrementNeck = true;
      }
      else
      {
        newNeckAngle = (incrementNeck) ? 163 : 2;
      }
      lastNeckUpdate = millis();
    }
    else if(newNeckAngle != neckAngle)
    {
      //Need to move neck
      if(millis() - lastNeckTurn > 10)
      {
        neckAngle = (newNeckAngle > neckAngle) ? neckAngle + 1 : neckAngle - 1;
        pwm.setPWM(NECK_SERVO, 0, pwToAnalogAng(neckAngle));
        lastNeckTurn = millis();
      }
    }
    oldNeckAngle = neckAngle;
  }
}

void roarJaws()
{
  if(jawsRoar)
  {
    int updateInterval = (newJawAngle == 116) ? 1500 : 2000;
    if(millis() - lastJawsUpdate > updateInterval && oldJawAngle == newJawAngle)
    {
      //Time to change neck direction
      newJawAngle = (oldJawAngle == 92) ? 116 : 92;
      Serial.println("JawAngleChange: " + String(newJawAngle));
      lastJawsUpdate = millis();
    }
    else if(newJawAngle != jawAngle)
    {
      //Need to move neck
      if(millis() - lastJawMove > 10)
      {
        Serial.println("JawAngleB: " + String(jawAngle));
        jawAngle = (newJawAngle > jawAngle) ? jawAngle + 1 : jawAngle - 1;
        Serial.println("JawAngleA: " + String(jawAngle));
        pwm.setPWM(JAWS_SERVO, 0, pwToAnalogAng(jawAngle));
        lastJawMove = millis();
      }
    }
    oldJawAngle = jawAngle;
  }
}

void wagTail()
{
  if(tailWag)
  {
    if(tailAngle == 120)
    {
      incrementTail = false;
    }
    else if(tailAngle == 40)
    {
      incrementTail = true;
    }
    if(millis() - lastTailUpdate > 15)
    {
      //update tail angle when interval is due
      tailAngle = (incrementTail) ? tailAngle + 1: tailAngle - 1;
      pwm.setPWM(TAIL_SERVO, 0, pwToAnalogAng(tailAngle));
      lastTailUpdate = millis();
    }
  }
}

void setup()
{
  // put your setup code here, to run once:
  Serial.begin(115200);

  pwm.begin();
  pwm.setPWMFreq(FREQUENCY);

  pwm.setPWM(testServo, 0, pwToAnalogAng(testAngle));
  pwm.setPWM(JAWS_SERVO, 0, pwToAnalogAng(92));
  pwm.setPWM(NECK_SERVO, 0, pwToAnalogAng(78));
  pwm.setPWM(TAIL_SERVO, 0, pwToAnalogAng(80));
  pwm.setPWM(WINGS_SERVO, 0, pwToAnalogAng(176));

  // Debug console
  Blynk.setDeviceName("JJ_ESP32");
  Blynk.begin(auth);
}

BLYNK_WRITE(V0)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    testAngle += 1;
    if(testAngle < 0)
    {
      testAngle = 0;
    }
    else if(testAngle > 180)
    {
      testAngle = 180;
    }
    Serial.println("Angle: " + String(testAngle));
    pwm.setPWM(testServo, 0, pwToAnalogAng(testAngle));
  }
}

BLYNK_WRITE(V1)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    testAngle -= 1;
    if(testAngle < 0)
    {
      testAngle = 0;
    }
    else if(testAngle > 180)
    {
      testAngle = 180;
    }
    Serial.println("Angle: " + String(testAngle));
    pwm.setPWM(testServo, 0, pwToAnalogAng(testAngle));
  }
}

BLYNK_WRITE(V2)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    pwm.setPWM(testServo, 0, pwToAnalogAng(testAngle));
  }
}

BLYNK_WRITE(V3)
{
  int pinValue = param.asInt();
  if(pinValue > -1 && pinValue < 16)
  {
    testServo = pinValue;
  }
}

BLYNK_WRITE(V4)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    wingsFlap = true;
  }
  else
  {
    wingsFlap = false;
  }
}

BLYNK_WRITE(V5)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    tailWag = true;
  }
  else
  {
    tailWag = false;
  }
}

BLYNK_WRITE(V6)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    neckRotate = true;
  }
  else
  {
    neckRotate = false;
  }
}

BLYNK_WRITE(V7)
{
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    jawsRoar = true;
  }
  else
  {
    jawsRoar = false;
  }
}

BLYNK_WRITE(V8)
{
  //Home position
  int pinValue = param.asInt();
  if(pinValue > 0)
  {
    pwm.setPWM(JAWS_SERVO, 0, pwToAnalogAng(92));
    pwm.setPWM(NECK_SERVO, 0, pwToAnalogAng(78));
    pwm.setPWM(TAIL_SERVO, 0, pwToAnalogAng(80));
    pwm.setPWM(WINGS_SERVO, 0, pwToAnalogAng(176));
  }
}

void loop()
{
  Blynk.run();
  flapWings();
  wagTail();
  rotateNeck();
  roarJaws();
}
